# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for data handling."""

import os

import tensorflow as tf
from google import protobuf


def preview_tfrecord(filepath):
    """Pretty prints a single record from a tfrecord file."""
    dataset = tf.data.TFRecordDataset(os.path.expanduser(filepath))
    example = tf.train.Example()
    example.ParseFromString(next(iter(dataset)).numpy())
    formatted = protobuf.text_format.MessageToString(
        example, use_short_repeated_primitives=True
    )
    print(formatted)
